#!/usr/bin/python
# -*- coding: utf-8 -*-


import os
import csv
import math
import numpy as np


def testMarkovMatrix(mat):
    '''
    Tests that the matrix behaves according to Markov properties (all
        rows sum to 1).
    '''
    rowMkvSums = list(map(sum, mat))
    check = [math.isclose(i, 1) for i in rowMkvSums]
    if check.count(True) == len(mat):
        return True
    return False


def maxAlleleInLandscape(geneLandscape):
    genesMaxes = map(np.amax, geneLandscape)
    landscapeMax = max(genesMaxes)
    return landscapeMax


def autoGenerateGenotypesDictionary(genes, genotypes):
    '''
    Description:
        * This function takes the genes and genotypes lists, and auto-generates
            the dictionary for the aggregation of genes.
    In:
        * genes: list of the genes to aggregate
        * genotypes: list of strings containing the genotypes
    Out:
        * Aggregation dictionary
    Notes:
        * This is the most general case of the aggregation. For specific
            applications, other functions will need to be created.
    '''
    tempList = []
    for gene in genes:
        tempList.append(findGeneInGenotypesList(gene, genotypes))
    returnDict = {
        "genotypes": genes,
        "indices": tempList
    }
    return returnDict


def findGeneInGenotypesList(gene, genotypes):
    '''
    Description:
        * Returns a list of the repetitions of the allele in a list of
            genotypes strings
    In:
        * gene: character of the allele we are looking for
        * genotypes: list of the genotypes present in the experiment
    Out:
        * List containing the counts of the appearances of the gene in
            the genotypes list
    Notes:
        * This function is mostly designed to be used in a wrapper
            (autoGenerateGenotypesDictionary)
    '''
    repsList = []
    for i in range(0, len(genotypes)):
        count = genotypes[i].count(gene)
        if count > 0:
            repsList.append([i] * count)
    return flatten(repsList)


def flatten(lst):
    '''
    Description:
        * Recursive function to flatten a list to a level 1
    In:
        * Any list
    Out:
        * Flattened list
    Notes:
        * Not coded by me
    '''
    return sum(([x] if not isinstance(x, list) else flatten(x)
                for x in lst), [])


def generateAggregationDictionary(genotypesList, indicesList):
    '''
    Description:
        * Creates a dictionary to be used in the aggregation functions.
    In:
        * List of genotypes strings (ordered)
        * List of the indices to aggregate
            (ordered and with duplicate instances)
    Out:
        * Dictionary containing:
            genotypesList
            indices
    Notes:
        * Probably unnecessary, but here so that I don't forget the names
            of the dictionary labels.
    '''
    aggregationDictionary = {
        "genotypes": genotypesList,
        "indices": indicesList
    }
    return aggregationDictionary


def listDirectoriesInPath(path):
    """
    Description:
        * This auxiliary function returns the names of all the folders within
            a given path.
    In:
        * path: Directory address
    Out:
        * List of folder names
    Notes:
        * NA
    """
    file = os.listdir(path)
    folderNames = [
        name for name in file if os.path.isdir(os.path.join(path, name))
    ]
    return folderNames


def listDirectoriesWithPathWithinAPath(pathFilename):
    '''
    Description:
        * Returns a list of folder paths within a given path.
    In:
        * pathFilename: Path to search directories in
    Out:
        * List of folder paths
    Notes:
        * NA
    '''
    folders = listDirectoriesInPath(pathFilename)
    for i, folder in enumerate(folders):
        folders[i] = pathFilename + folder
    return folders


def getUniqueGenesFromGenotypes(
    genotypes
):
    """
    Description:
        * From a list of genotypes, returns the list of unique genes.
    In:
        * genotypes: List of genotypes present in the experiment.
    Out:
        * genesList: List of unique genes present.
    Notes:
        * Use in tandem with: readGenotypes
    """
    genesList = list({l for word in genotypes for l in word})
    return genesList


def readGenotypes(filename, skipColumns=1):
    '''
    Description:
        * Returns the genotypes header from an run file.
    In:
        * filename: Path to the run file
        * skipColumns: Optional argument for compatibility in the cases where
            the patch number is still present in the data.
    Out:
        * List of genotypes strings
    Notes:
        * NA
    '''
    reader = csv.reader(open(filename))
    return next(reader)[skipColumns:]


def splitExperimentString(experimentString):
    """
    Description:
        * Ad hoc auxiliary function to retrieve releases number and coverage
            from the folder name.
    In:
        * Experiment folder name.
    Out:
        * Dictionary containing:
            releasesNumber
            coverage
    Notes:
        * Not to be used in different experiments
    """
    split = experimentString.split("_")
    releasesNumber = int(split[3])
    coverage = int(split[4])
    return {"releasesNumber": releasesNumber, "coverage": coverage}


def calculateMaxPopInLandscapeReps(landscapeReps):
    """
    Description:
        * Returns the max population from a set of experiments in garbage
            folder.
    In:
        * landscapeReps: Dictionary with the info for the traces (garbage)
    Out:
        * maxPop: Maximum population found in the experiments.
    Notes:
        * Mainly developed to set the "y" axis range in traces plots.
    """
    landscapes = landscapeReps["landscapes"]
    list = [None] * len(landscapeReps["landscapes"][0])
    for i in range(len(landscapeReps["landscapes"][0])):
        list[i] = sum(landscapes[0][i][0])
    return max(list)


def getGenotypesWithPositions(genotypes, allelePositions):
    """
    Description:
        * Returns a list of indices corresponding to genotypes with a
            desired allelePosition
    In:
        * genotypes: List of genotype strings
        * allelePositions: List of (char, int) tuples (allele, index in
            genotype string)
    Out:
        * indices: Indices of strings in genotypes that have an allele in a
            specified position
    Notes:
        * Indices are repeated if they have multiple alleles in specified
            positions
    """
    indices = []
    for i in range(len(genotypes)):
        s = genotypes[i]
        for allele, index in allelePositions:
            if s[index] == allele:
                indices.append(i)
    return indices


def quickSaveTraceAggData(
    aggData,
    filename,
    fmt="%.10d"
):
    np.savetxt(
        filename,
        aggData["population"],
        header=(",".join(aggData["genotypes"])),
        delimiter=",",
        fmt=fmt,
        comments=''
    )


def quickSaveRepsAggData(
    landscapeReps,
    foldername,
    fmt="%.10d",
    padNumb=5
):
    if not os.path.exists(foldername):
        try:
            os.mkdir(foldername)
        except:
            raise OSError("Can't create destination directory (%s)!" %
                          (foldername))

    repsNumber = len(landscapeReps["landscapes"])
    for i in range(0, repsNumber):
        nodesNumber = len(landscapeReps["landscapes"][0])
        for j in range(0, nodesNumber):
            aggData = {
                "genotypes": landscapeReps["genotypes"],
                "population": (landscapeReps["landscapes"][i][j])
            }
            quickSaveTraceAggData(
                aggData,
                foldername + "/N" + str(j).rjust(padNumb, "0") +
                "_R" + str(i).rjust(padNumb, "0") + ".csv",
                fmt=fmt
            )



from .auxiliaryFunctions import *
from .colors import *
from .experimentsHandler import *
from .factorialExperimentsHandler import *
from .factorialSensitivity import *
from .plots import *
from .summaryStatistics import *
from .tracesHandler import *
from .spatialVideos import *
from .landscape import *




import glob
import matplotlib.pyplot as plt
import MoNeT_MGDrivE as monet
from mpl_toolkits.basemap import Basemap
import numpy as np
import subprocess


def getClusters(fileLocation):
    coordinates = [[], []]
    clusterFile = open(fileLocation, 'r')
    for line in clusterFile:
        tokens = line.split(',')
        coordinates[0].append(float(tokens[0]))
        coordinates[1].append(float(tokens[1]))
    return coordinates


def populateClusters(clusterNum, cFileName, pFileLocation, pFilePattern={}):

    clusters = []
    for i in range(clusterNum):
        clusters.append({'male': [], 'female': []})
    patchCluster = []
    coordFile = open(cFileName, 'r')

    for line in coordFile:
        patchCluster.append(int(line.split(',')[2].strip()))

    if 'male' in pFilePattern:
        patchFileList = sorted(glob.glob(pFileLocation+pFilePattern['male']))
    else:
        patchFileList = sorted(glob.glob(pFileLocation+'/ADM_*'))
        if not patchFileList:
            patchFileList = sorted(glob.glob(pFileLocation+'/M_*'))

    for index, patchFileN in enumerate(patchFileList):
        clusters[patchCluster[index]]['male'].append(patchFileN)

    if 'female' in pFilePattern:
        patchFileList = sorted(glob.glob(pFileLocation+pFilePattern['female']))
    else:
        patchFileList = sorted(glob.glob(pFileLocation+'/AF1_*'))
        if not patchFileList:
            patchFileList = sorted(glob.glob(pFileLocation+'/F_*'))

    for index, patchFileN in enumerate(patchFileList):
        clusters[patchCluster[index]]['female'].append(patchFileN)

    return clusters


def aggregateClusters(clusters, aggDict):
    aggList = []
    for fileList in clusters:
        aggPatches = monet.loadAndAggregateLandscapeData({
            'male': fileList['male'],
            'female': fileList['female']},
            aggDict,
            male=True, female=True
            )["landscape"]
        res = np.zeros_like(aggPatches[0])
        for patch in aggPatches:
            res += patch
        aggList.append(res)
    return aggList


def getGenotypes(fileName):
    f = open(fileName, 'r')
    header = next(f).split(',')[1:]
    f.close()
    return header


def createFig(coordinates):
    fig = None
    ax = None
    m = None
    minLat = min(coordinates[1])
    maxLat = max(coordinates[1])
    minLong = min(coordinates[0])
    maxLong = max(coordinates[0])
    fig, ax = plt.subplots(figsize=(5, 5))
    ax.tick_params(
            axis='both',          # changes apply to the both
            which='both',      # both major and minor ticks are affected
            bottom=False,      # ticks along the bottom edge are off
            top=False,         # ticks along the top edge are off
            left=False,
            right=False,
            labelbottom=False,  # labels along the bottom edge are off
            labelleft=False
        )
    m = Basemap(
            projection='merc', llcrnrlat=minLat-0.001, urcrnrlat=maxLat+0.001,
            llcrnrlon=minLong-0.001, urcrnrlon=maxLong+0.001, lat_ts=20,
            resolution='i', ax=ax
        )
    m.drawcounties(linewidth=0.3)
    return (fig, ax, m)


def getAlphas(popList):
    total = np.sum(popList)
    return ([float(x)/total for x in popList], total)


def draw_dots(m, alphas, colorList, long=0, lat=0, size=60):
    # start = 0.0
    for idx, value in enumerate(alphas):
        m.scatter([long], [lat], latlon=True, marker=(6, 0), s=max(
            9, 0.15*size), facecolor=colorList[idx], alpha=value)


def generateClusterGraphs(aggList, coordinates, destination, colorList):
    time = len(aggList[0])
    fig, ax, m = createFig(coordinates)
    for tick in range(time):
        for idx, cData in enumerate(aggList):
            pops = []
            try:
                pops = cData[tick]
                alphas, size = getAlphas(pops)
                if alphas:
                    draw_dots(m, alphas, colorList,
                              coordinates[0][idx], coordinates[1][idx], size)
                else:
                    continue
            except Exception as e:
                return e
        else:
            fig.savefig(destination+'/c_'+str(tick).zfill(6)+".png",
                        dpi=512, orientation='portrait', papertype=None,
                        transparent=True, format="png",
                        bbox_inches='tight', pad_inches=0.05, frameon=None)
            plt.close(fig)
            plt.close('all')
            fig, ax, m = createFig(coordinates)
    return


def generateVideo(name, background, imageLocation, imagePattern):
    video = subprocess.Popen([
            'ffmpeg', '-i', background, '-r', '24', '-f',
            'image2', '-s', '1920x1080', '-i', imageLocation+imagePattern,
            '-vcodec', 'libx264', '-crf', '25', '-filter_complex',
            '[0:v][1:v]overlay=0:0[bg]; [bg]pad=ceil(iw/2)*2:ceil(ih/2)*2, '
            + 'format=yuv420p[outv]', '-map', '[outv]', name]
        )
    print(video.pid)
    return video



import glob
import numpy as np
import MoNeT_MGDrivE.auxiliaryFunctions as auxFun


def readExperimentFilenamesEpi(
    experimentPath,
    stateIdentifiers={
        "human": "HUM",
        "male": "ADM",
        "e_female": "E_AF1",
        "i_female": "I_AF1", "s_female": "S_AF1"
    }
):
    """
    Description:
        * This auxiliary function searches within a given path, and returns all
            the CSV files that match the head provided by the dictionary keys
            for human, male, and infected state of female
    In:
        * experimentPath: Path to the directory that contains the experiments'
            CSV files (quantiles over patches).
        * stateIdentifiers: Dictionary containing the head identifier for
            the human, male, and state of female CSV files.
    Out:
        * Dictionary with the sorted lists of filenames associated with each
            file head given in stateIdentifiers
            ex.  "human" [list -> strings], "male" [list -> strings], etc

    """
    stateIdToFiles = {}
    for stateId in stateIdentifiers:
        files = sorted(
            glob.glob(
                experimentPath + "/" + stateIdentifiers[stateId] + "*.csv"
            )
        )
        stateIdToFiles[stateId] = files

    return stateIdToFiles


def loadNodeDataEpi(
    filename,
    dataType=float,
    skipHeader=1,
    skipColumns=1
):
    """
    Description:
        * Loads the data given by filename
    In:
        * filename: path to csv file
        * dataType: To save memory/processing time if possible (int/float).
    Out:
        * Dictionary containing:
            "genotypes" [list -> strings]
            "population" [numpyArray]

    """
    genotypes = auxFun.readGenotypes(filename)
    data = np.genfromtxt(
            filename,
            dtype=dataType,
            skip_header=skipHeader,
            delimiter=",")
    returnDictionary = {
            "genotypes": genotypes,
            "population": data[:, skipColumns:]
        }
    return returnDictionary


def loadLandscapeDataEpi(
    filenames,
    male=True,
    female=True,
    dataType=float
):
    """
    Description:
        * Imports the information of all the nodes in filenames
    In:
        * filenames: List of paths to desired csv files
    Out:
        * Dictionary containing:
            "genotypes" [list -> strings]
            "landscape" [list -> numpyArrays]
    """
    if len(filenames) == 0:
        return
    genotypes = auxFun.readGenotypes(filenames[0])
    nodesDataList = []
    for f in filenames:
        data = loadNodeDataEpi(f, dataType)["population"]
        nodesDataList.append(data)

    returnDictionary = {
        "genotypes": genotypes,
        "landscape": nodesDataList
    }
    return returnDictionary


def sumLandscapePopulationsEpi(
    landscapeData
):
    """
    Description:
        * This function sums the data in each np array given by
            landscapeData["landscape"]
    In:
        * landscapeData: Data loaded with the "loadLandscapeDataEpi" function.
    Out:
        * Dictionary containing:
            "genotypes" [list -> strings]
            "population" [numpyArray]
    """
    dataList = landscapeData["landscape"]
    if len(dataList) == 0:
        return

    fillArray = np.zeros_like(dataList[0])
    for i in range(len(dataList)):
        fillArray += dataList[i]

    returnDictionary = {
        "genotypes": landscapeData["genotypes"],
        "population": fillArray
    }
    return returnDictionary


def sumAlleleCounts(sumLandscapeData, alleleNames, columns):
    """
    Description:
        * This function sums the total count of each allele given by
            sumLandscapeData["population"]
    In:
        * sumLandscapeData: Dict loaded with the "sumLandscapePopulationsEpi"
            function.
        * alleleNames: list of individual alleles, eg ["W", "R"]
        * columns: List of lists describing number of times each one
            indexed column should be counted for this allele
            ex. [[1, 1, 2], [2, 3, 3]] if the given genotypes are WW, WR,
            and RR
    Out:
        * Dictionary containing:
            "genotypes" [list -> strings]
            "population" [numpyArray]
    """
    landscapeData = sumLandscapeData["population"]
    fillArray = np.zeros((len(landscapeData), len(alleleNames)))

    for i in range(len(alleleNames)):
        for index in columns[i]:
            # subtract 1 because index is 1 indexed
            fillArray[:, i] += landscapeData[:, index-1]

    returnDictionary = {
        "alleles": alleleNames,
        "totalCounts": fillArray
    }
    return returnDictionary



import math
import numpy as np
import scipy.stats as stats
#import vincenty as vn

# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# Constants
# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
AEDES_EXP_PARAMS = [0.01848777, 1.0e-10, math.inf]

# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# Distances
# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

def euclideanDistance(a, b):
    '''
    Calculates the Euclidean distance between two-dimensional coordinates.
    '''
    dist = math.sqrt((a[0] - b[0]) ** 2 + (a[1] - b[1]) ** 2)
    return dist


def calculateDistanceMatrix(landscape, distFun=euclideanDistance):
    '''
    Returns the distance matrix according to the provided distance function.
        Examples of these are: euclideanDistance (xy), vn.vincenty (latlong).
    '''
    coordsNum = len(landscape)
    distMatrix = np.empty((coordsNum, coordsNum))
    for (i, coordA) in enumerate(landscape):
        for (j, coordB) in enumerate(landscape):
            distMatrix[i][j] = distFun(coordA, coordB)
    return distMatrix


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#  Linear Kernel Functions
# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

def inverseLinearStep(distance, params=[.75, 1]):
    '''
    This function returns a migration estimate based on the inverse of the
        distance. NOTE: This is a terrible way to do it, but it's a first
        approximation. Should be replaced with the zero-inflated exponential.
    '''
    if math.isclose(distance, 0):
        return params[0]
    else:
        return (1 / (distance * params[1]))
    return True


def zeroInflatedLinearMigrationKernel(
            distMat,
            params=[.75, 1]
        ):
    '''
    Takes in the distances matrix, zero inflated value (step) and two extra
        parameters to determine the change from distances into distance-based
        migration probabilities (based on the kernel function provided).
    '''
    coordsNum = len(distMat)
    migrMat = np.empty((coordsNum, coordsNum))
    for (i, row) in enumerate(distMat):
        for (j, dst) in enumerate(row):
            migrMat[i][j] = inverseLinearStep(dst, params=params)
        # Normalize rows to sum 1
        migrMat[i] = migrMat[i] / sum(migrMat[i])
    return migrMat


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# Exponential Migration Kernels
# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

def truncatedExponential(distance, params=AEDES_EXP_PARAMS):
    '''
    Calculates the zero-inflated exponential for the mosquito movement kernel
        (default parameters set to Aedes aegypti calibrations).

        params = [rate, a, b]
    '''
    if(params[1] > params[2]):
        return None

    scale = 1.0/params[0]
    gA = stats.expon.cdf(params[1], scale=scale)
    gB = stats.expon.cdf(params[2], scale=scale)
    if np.isclose(gA, gB):
        return None

    densNum = stats.expon.pdf(distance, scale=scale)
    densDen = gB - gA

    return densNum/densDen


def zeroInflatedExponentialMigrationKernel(
            distMat,
            params=AEDES_EXP_PARAMS,
            zeroInflation=.75
        ):
    coordsNum = len(distMat)
    migrMat = np.empty((coordsNum, coordsNum))
    for (i, row) in enumerate(distMat):
        for (j, dst) in enumerate(row):
            if(i == j):
                migrMat[i][j] = 0
            else:
                migrMat[i][j] = truncatedExponential(dst, params=params)
        migrMat[i] = migrMat[i] / np.sum(migrMat[i]) * (1 - zeroInflation)
    np.fill_diagonal(migrMat, zeroInflation)
    return migrMat



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='MoNeT_MGDrivE',
    version='0.5.3',
    url='https://github.com/Chipdelmal/MoNeT_MGDrivE',
    author='Hector M. Sanchez C.',
    author_email='sanchez.hmsc@berkeley.edu',
    description="MoNeT python package for MGDrivE data analysis",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=[
        'numpy', 'scipy', 'matplotlib', 'ipython',
        'jupyter', 'pandas', 'sympy', 'vincenty'
    ],
    license='MIT',
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Operating System :: OS Independent",
    ]
 )



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import MoNeT_MGDrivE.colors as monetPlots


def plotMeanGenotypeTrace(aggData, style):
    """
    Description:
        * Plots the mean response of an aggregate dataset.
    In:
        * aggData: dictionary containing "genotype" and "populations" pairs
        * style: dictionary containing width, colors, aspect, alpha
    Out:
        * fig: matplotlib figure
    Notes:
        * NA
    """
    groups = aggData['genotypes']
    pops = aggData['population']
    time = np.arange(len(pops))
    df = pd.DataFrame(time, columns=['Time'])
    final = [df[['Time']] for _ in range(len(groups))]
    local = pd.DataFrame(pops, columns=groups)
    fig, ax = plt.subplots()
    ax.set_aspect(aspect=style["aspect"])
    # plt.xticks([])
    # plt.yticks([])
    for j in range(len(groups)):
        final[j].insert(1, groups[j] + str(1), (local[groups[j]]).copy())
        final[j] = final[j].set_index('Time')
    for i in range(len(final)):
        final[i].plot(
            ax=ax, linewidth=style["width"], legend=False,
            color=style["colors"][i], alpha=style["alpha"]
        )
    legends = []
    for i in range(len(groups)):
        legends.append(
            mpatches.Patch(color=style["colors"][i], label=groups[i])
        )
    if style["legend"] is True:
        plt.legend(bbox_to_anchor=(1.05, 1), loc=2,
                   ncol=2, borderaxespad=0.)
    ax.xaxis.set_label_text("")
    ax.yaxis.set_label_text("")
    # plt.ylabel("Allele Count")
    return fig


def plotMeanGenotypeStack(
    aggData,
    style,
    vLinesCoords=[]
):
    """
    Description:
        * Plots the mean response of an aggregate dataset.
    In:
        * aggData: dictionary containing "genotype" and "populations" pairs
        * style: dictionary containing width, colors, aspect, alpha
    Out:
        * fig: matplotlib figure
    Notes:
        * NA
    """
    groups = aggData['genotypes']
    pops = aggData['population']
    time = np.arange(len(pops))
    df = pd.DataFrame(time, columns=['Time'])
    final = [df[['Time']] for _ in range(len(groups))]
    local = pd.DataFrame(pops, columns=groups)
    fig, ax2 = plt.subplots()
    ax2.set_aspect(aspect=style["aspect"])
    allele_dict = {}
    for j in range(len(groups)):
        final[j].insert(1, groups[j] + str(1), (local[groups[j]]).copy())
        final[j] = final[j].set_index('Time')
    for i in range(len(groups)):
        allele_dict[groups[i]] = final[i].T.sum()
    res = pd.DataFrame(allele_dict)
    res = res.reindex(columns=groups)
    res.plot(
        kind='area', ax=ax2, legend=style["legend"], color=style["colors"],
        linewidth=style["width"], alpha=style["alpha"]
    )
    for i in range(len(vLinesCoords)):
        ax2.plot(
            [vLinesCoords[i], vLinesCoords[i]],
            [style["yRange"][0], style["yRange"][1]],
            'k--',
            lw=.5
        )
    plt.ylabel("")
    plt.xlabel("")
    if style["legend"] is True:
        plt.legend(bbox_to_anchor=(1.05, 1), loc=2,
                   ncol=2,  borderaxespad=0.)
    return fig


def plotGenotypeFromLandscape(
    landscapeGene,
    style={"aspect": 12, "cmap": monetPlots.cmaps[0]}
):
    """
    Description:
        * Creates the heatmap plot of a gene array
    In:
        * landscapeGene: spatiotemporal array for the gene to plot
        * style: styling options for the plot
    Out:
        * fig: matplotlib figure
    Notes:
        * NA
    """
    fig, ax = plt.subplots(nrows=1, figsize=(20, 5))
    ax.imshow(landscapeGene, cmap=style["cmap"], aspect=style["aspect"])
    return fig


def plotGenotypeArrayFromLandscape(
    geneSpatiotemporals,
    style={"aspect": 12, "cmap": monetPlots.cmaps}
):
    """
    Description:
        * Creates the heatmap plot of all the genotypes in the landscape
            separately.
    In:
        * landscapeData: population dynamics data
        * style: styling options for the plot
    Out:
        * fig: matplotlib figure
    Notes:
        * NA
    """
    genesNumber = len(geneSpatiotemporals["genotypes"])
    plotsList = [None] * genesNumber
    for i in range(0, genesNumber):
        plotsList[i] = plotGenotypeFromLandscape(
            geneSpatiotemporals["geneLandscape"][i],
            style={"aspect": style["aspect"], "cmap": style["cmap"][i]}
        )
    plotsDict = {
        "genotypes": geneSpatiotemporals["genotypes"],
        "plots": plotsList
    }
    return plotsDict


def plotGenotypeOverlayFromLandscape(
    geneSpatiotemporals,
    style={"aspect": 12, "cmap": monetPlots.cmaps},
    vmin=None,
    vmax=None
):
    """
    Description:
        * Plots the combined "landscape-heatmap" plots in one.
    In:
        * geneSpatiotemporals: Array of the spatiotemporal genotypes
            information (gene counts across nodes through time).
        * style: styling options for the plot
        * vmin:
        * vmax:
    Out:
        * fig: matplot fig of combined heatmaps
    Notes:
        * NA
    """
    alleleNames = geneSpatiotemporals["genotypes"]
    counts = geneSpatiotemporals["geneLandscape"]
    fig = plt.figure(figsize=(20, 5))
    for i in range(len(alleleNames)):
        plt.imshow(
            counts[i],
            cmap=style["cmap"][i],
            aspect=style["aspect"],
            vmin=vmin, vmax=vmax
        )
    return fig


def plotNodeDataRepetitions(
    nodeRepetitionsArray,
    style
):
    """
    Description:
        * Generates the "traces" plot for one node.
    In:
        * nodeRepetitionsArray: Intermediate structure generated by taking
            the information of a given node accross all landscapes.
        * style: styling options for the plot
    Out:
        * fig: matplotlib traces figure
    Notes:
        * This function is meant to work within plotLandscapeDataRepetitions,
            so it's not necessary to call it directly.
    """
    probeNode = nodeRepetitionsArray
    repsNumber = len(probeNode)
    genesNumber = len(probeNode[0][0])
    fig, ax = plt.subplots()
    ax.set_aspect(aspect=style["aspect"])
    for j in range(0, repsNumber):
        transposed = probeNode[j].T
        for gene in range(0, genesNumber):
            ax.plot(
                transposed[gene],
                linewidth=style["width"],
                color=style["colors"][gene],
                alpha=style["alpha"]
            )
    return fig


def plotLandscapeDataRepetitions(
    landscapeReps,
    style
):
    """
    Description:
        * Generates the individual "traces" plots for a whole landscape.
    In:
        * landscapeReps: landscape repetitions data generated with
            loadAndAggregateLandscapeDataRepetitions.
        * style: styling options for the plot.
    Out:
        * figs: array of matplotlib traces figures.
    Notes:
        * NA
    """
    landscapes = landscapeReps["landscapes"]
    landscapesNumb = len(landscapeReps["landscapes"][0])
    figs = [None] * landscapesNumb
    for i in range(0, landscapesNumb):
        probeNode = list(zip(*landscapes))[i]
        figs[i] = plotNodeDataRepetitions(probeNode, style)
    return figs


def plotAllTraces(
    landscapeReps,
    style
):
    """
    Description:
        * Generates the individual "traces" plot for a whole landscape.
    In:
        * landscapeReps: landscape repetitions data generated with
            loadAndAggregateLandscapeDataRepetitions.
        * style: styling options for the plot.
    Out:
        * fig: a matplotlib traces figure with all of the information on the landscapeReps.
    Notes:
        * NA
    """
    repetitions = len(landscapeReps["landscapes"])
    nodesNumb = len(landscapeReps["landscapes"][0])
    genesNumber = len(landscapeReps["landscapes"][0][0][0])
    fig, ax = plt.subplots()
    ax.set_aspect(aspect=style["aspect"])
    for rep in landscapeReps["landscapes"]:
        for node in rep:
            transposed = node.T
            for gene in range(0, genesNumber):
                ax.plot(
                    transposed[gene],
                    linewidth=style["width"],
                    color=style["colors"][gene],
                    alpha=style["alpha"]
                )

    return fig


def quickSaveFigure(
    fig,
    path,
    dpi=1024,
    format="png"
):
    """
    Description:
        * Standardized method to save experiments figures.
    In:
        * fig: figure to save
        * path: directory to save to
        * dpi: resolution
        * format: image format
    Out:
        * NA: (saves to disk)
    Notes:
        * NA
    """
    fig.savefig(
        path, dpi=dpi, facecolor='w',
        edgecolor='w', orientation='portrait', papertype=None,
        format=format, transparent=True, bbox_inches='tight',
        pad_inches=0, frameon=None
    )





def reachedSteadtStateAtDay(
    aggData,
    safety=.01,
    finalFrame=-1
):
    """
    Description:
        * Calculates the point at which the aggregated alleles reach their
            steady state (defined by the final frame of the simulation).
    In:
        * aggData: Genotypes aggregated data.
        * safety: Envelope of values around the steady state that are
            considered "stable" (as a proportion of the final total allele
            composition).
        * finalFrame: Index of the day to be considered as "stable" for
            reference purposes.
    Out:
        * steadtStateReach: Day at which the dynamics of the system became
            stable.
    Notes:
        * This is a quick and simple way to calculate the steady state, but
            assumes that the system stabilizes and that the simulation was
            run for enough time for the dynamics to become stable.
    """
    finalFrame = aggData["population"][finalFrame]
    tolerance = round(sum(finalFrame) * safety)
    toleranceUp = finalFrame + tolerance
    toleranceDown = finalFrame - tolerance

    daysMax = len(aggData["population"])

    for i in range(0, daysMax):
        steadyStateReach = daysMax
        testFrame = aggData["population"][i]

        boolsUp = testFrame < toleranceUp
        boolsDown = testFrame > toleranceDown

        if all(boolsUp) and all(boolsDown):
            steadyStateReach = i
            break

    return steadyStateReach



from matplotlib.colors import LinearSegmentedColormap

# Define 5 colormaps ranging from transparent to opaque.
cdict1 = {
    'red':   ((0.0, 1.0, 1.0), (1.0, 1.0, 1.0)),
    'green': ((0.0, 0.0, 0.0), (1.0, 0.0, 0.0)),
    'blue':  ((0.0, 0.0, 0.0), (1.0, 0.3, 0.3)),
    'alpha': ((0.0, 0.0, 0.0), (1.0, 1.0, 1.0)),
}
red1 = LinearSegmentedColormap('Red1', cdict1)

cdict2 = {
    'red':   ((0.0, 0.3, 0.3), (1.0, 0.3, 0.3)),
    'green': ((0.0, 0.5, 0.5), (1.0, 0.5, 0.5)),
    'blue':  ((0.0, 1.0, 1.0), (1.0, 1.0, 1.0)),
    'alpha': ((0.0, 0.0, 0.0), (1.0, 1.0, 1.0)),
}
light_blue1 = LinearSegmentedColormap('LightBlue1', cdict2)

cdict3 = {
    'red':   ((0.0, 0.5, 0.5), (1.0, 0.5, 0.5)),
    'green': ((0.0, 0.0, 0.0), (1.0, 0.0, 0.0)),
    'blue':  ((0.0, 1.0, 1.0), (1.0, 1.0, 1.0)),
    'alpha': ((0.0, 0.0, 0.0), (1.0, 1.0, 1.0)),
}
purple1 = LinearSegmentedColormap('Purple1', cdict3)

cdict4 = {
    'red':   ((0.0, 1.0, 1.0), (1.0, 1.0, 1.0)),
    'green': ((0.0, 0.0, 0.0), (1.0, 0.0, 0.0)),
    'blue':  ((0.0, 1.0, 1.0), (1.0, 1.0, 1.0)),
    'alpha': ((0.0, 0.0, 0.0), (1.0, 1.0, 1.0)),
}
pink1 = LinearSegmentedColormap('Pink1', cdict4)

cdict5 = {
    'red':   ((0.0, 0.0, 0.0), (1.0, 0.0, 0.0)),
    'green': ((0.0, 0.25, 0.25), (1.0, 0.25, 0.25)),
    'blue':  ((0.0, 0.75, 0.75), (1.0, 0.75, 0.75)),
    'alpha': ((0.0, 0.0, 0.0), (1.0, 1.0, 1.0)),
}
dark_blue1 = LinearSegmentedColormap('DarkBlue1', cdict5)

rgba_colors = [
    (1, 0, 0.3, 0.7), (1, 0, 1, 0.7),
    (0.5, 0, 1, 0.7), (0, 0.325, 0.75, 0.7)
]
cmaps = [light_blue1, red1, purple1, pink1, dark_blue1]


def generateAlphaColorMapFromColor(
    color
):
    """
    Description:
        * Generates the cmap necessary to do the landscape heatmaps.
    In:
        * color: hex color to alpha cmap
    Out:
        * cmap: Color map changing in alpha
    Notes:
        * NA
    """
    alphaMap = LinearSegmentedColormap.from_list(
        'tempMap',
        [(0.0, 0.0, 0.0, 0.0), color],
        gamma=0
    )
    return alphaMap


def generateAlphaColorMapFromColorArray(
    colorArray
):
    """
    Description:
        * Generates a list of cmaps from a colors list.
    In:
        * colorArray: Array containing the colors to cmap.
    Out:
        * cmapsList: List of cmaps
    Notes:
        * NA
    """
    elementsNumb = len(colorArray)
    cmapsList = [None] * elementsNumb
    for i in range(0, elementsNumb):
        cmapsList[i] = generateAlphaColorMapFromColor(colorArray[i])
    return cmapsList
